-- lib_runflex.lua
-- creates cages from very compact definitions
-- (see examples)

package.path = package.path..";../?.lua"
require "lib_safelua"
require "lib_list"

blockingOp = nil

function makeSuperpower(name, cage)
  if name == 'socket' then
    require "socket"
    return socket
  end
  
  if name == 'vcl' then
    require "vcl"
    return VCL
  end

  if name == 'blockingOp' then  
    return function(f)
      blockingOp = f
    end
  end

  if name == 'cageManagement' then
    local cageManagement = {}

    function cageManagement:sandboxIds()
      local keys = {}
      for id, _ in pairs(cage.sandboxInfos) do
        table.insert(keys, id)
      end
      return keys
    end

    function cageManagement:getInfo(id)
      local info = cage.sandboxInfos[id]
      local routine = cage:findRoutine(id)
      if info.error then
        info.status = 'error'
      elseif routine == nil then
        info.status = 'done'
      else
        info.status = routine.sleeping and 'sleeping' or 'running'
      end
      return info
    end

    function cageManagement:startSandbox(def)
      --print "adding sandbox"
      id = addSandbox(cage, def)
      --print("sandbox added, #="..tostring(#cage.sandboxes))
      makeCoroutine(cage, #cage.sandboxes)
      return id
    end

    function cageManagement:deleteSandbox(id)
      local info = cage.sandboxInfos[id]
      if info == nil then return end
      cage.sandboxInfos[id] = nil
      if info.shutdownHandler then
        info.shutdownHandler()
      end

      -- TODO: delete safecomm queue, channels
      for i=1, #cage.routines do
        if cage.routines[i].id == id then
          table.remove(cage.routines, i)
          break
        end
      end
    end

    function cageManagement:sandboxCode(id)
      local info = cage.sandboxInfos[id]
      if info then return info.code end
    end

    function cageManagement:sandboxOutput(id)
      local info = cage.sandboxInfos[id]
      if info then
        local lines = info.output
        if info.error then lines = list.concat(lines, {info.error}) end
        return table.concat(lines, '\n')
      end
    end

    return cageManagement
  end
  
  error("Unknown type of magic: "..name)
end

-- define a sandbox
function Sandbox(data)
  return data
end

function printtostring(...)
  l = {}
  for _, x in ipairs(arg) do
    table.insert(l, tostring(x))
  end
  return table.concat(l, ' ') -- use space instead of tab
end

function addSandbox(cage, sandboxdef)
  -- either code or script must be set
  local code = sandboxdef.code
  if code then code = "** "..code else code = sandboxdef.script end
  table.insert(cage.filenames, code)

  cage.idCounter = cage.idCounter+1
  local id = 'sandbox'..cage.idCounter
  local sandbox = makeSandbox(id, cage)
  local info = cage.sandboxInfos[id]
  info.script = sandboxdef.code and "<inline>" or sandboxdef.script
  info.timeout = sandboxdef.timeout
  info.softtimeout = sandboxdef.softtimeout

  -- replace 'print'
  sandbox.print = function (...)
    local s = printtostring(...)
    print("["..id.."]", s)
    table.insert(info.output, s)
  end

  -- add safecomm
  addSafecomm(sandbox, cage)

  -- add magic powers - if selected for this sandbox in config
  local powers = sandboxdef.magic or {}
  info.magic = powers
  function sandbox.safelua:requestExtension(name)
    for power in elems(powers) do
      if power == name then
        return makeSuperpower(name, cage)
      end
    end

    error("Extension unavailable: "..name)
  end

  table.insert(cage.sandboxes, sandbox)
  return id
end

-- construct and start a new cage from table definition
function Cage(data)
  cage = makeCage()
  cage.timeout = data.timeout
  cage.softtimeout = data.softtimeout
  cage.sandboxes = {}
  cage.filenames = {}
  for _, sandboxdef in ipairs(data) do
    addSandbox(cage, sandboxdef)
  end
  
  -- follows: a nasty hack for VCLua
  local m = data[1].magic
  if m ~= nil and m[1] == 'vcl' then
    safecomm = cage:makeSafecommInterface('guievents')
    function onClick(Sender)
      print('global onclick '..tostring(Sender.Name))
      --[[for _, v in ipairs(theRoutines) do
        print(v.id..": "..coroutine.status(v.routine))
      end]]
      i = Sender.Name:find('_')
      if i then
        safecomm:send(Sender.Name:sub(1, i-1), Sender.Name:sub(i+1))
        doStuff()
        -- print "yielding"
        -- coroutine.yield(true)
      end
    end
  end
  
  runCoroutines(cage)
  doStuff()

  -- follows: another very nasty hack for VCLua
  -- (that doesn't seem to do the trick either)
  if blockingOp ~= nil then
    for _, v in ipairs(theRoutines) do
      print(v.id..": "..coroutine.status(v.routine))
    end
    local op = blockingOp
    blockingOp = nil
    -- r = coroutine.create(op)
    -- coroutine.resume(r)
    op()
  end
end

