-- lib_safecomm.lua
-- Implements the "safecomm" facilities for inter-sandbox communication
--
-- Version: Safe Lua 0.2 (http://safelua.sf.net)
-- by Stefan Reich (www.friedensrevolution.de)
-- Contact: stefan.reich.maker.of.eye@gmail.com
-- This code is public domain!

require "lib_datadumper"

-- from stdlib
--- An iterator over the elements of a list.
-- @param l list to iterate over
-- @return iterator function which returns successive elements of the list
-- @return the list <code>l</code> as above
-- @return <code>true</code>
function elems (l)
  local n = 0
  return function (l)
    n = n + 1
    if n <= #l then
      return l[n]
    end
   end,
  l, true
end

-- TODO: this logic does not seem to allow something like
-- send(sender, 'answer', {answerId=data._answerId, answer=answer})
-- (or does it?)
-- maybe drop this and stick with a standard form
function safecomm_extract(...)
  local recp, cmd, data
  if #arg == 1 then
    -- send called with { } syntax
    local tbl = arg[1]
    recp, cmd = tbl[1], tbl[2]
    data = tbl
    table.remove(data, 1)
    table.remove(data, 1)
  else
    -- send called with ( ) syntax
    recp, cmd = arg[1], arg[2]
    --dump(arg)
    data = arg
    table.remove(data, 1)
    table.remove(data, 1)
    data.n = nil
  end
  return recp, cmd, data 
end

-- A cage is a collection of sandboxes which may communicate through safecomm
-- and are run simultaneously, as coroutines (they call it cooperative
-- multitasking)
function makeCage()
  local cage = {}
  
  local safecomm_queues = {}
  local safecomm_channels = {}
  local safecomm_keepAlive = {}
  cage.safecomm_keepAlive = safecomm_keepAlive
  cage.sandboxInfos = {}
  cage.idCounter = 0

  function cage:makeSafecommInterface(sandboxId)
    local safecomm_interface = {}
    local handlers = {}
    
    function safecomm_interface:send(...)
      local recp, cmd, data = safecomm_extract(...)
      print("safecomm from "..sandboxId.." to "..recp..", cmd: "..cmd.." "..tostr(data))
      
      -- new method
      local c = safecomm_channels[recp]
      if c ~= nil then
        -- print("Channel found ("..#c..")")
        --dump(c)
        for _, id in ipairs(c) do
          --print("Channel found: "..recp)
          safecomm_queues[id] = safecomm_queues[id] or {}
          table.insert(safecomm_queues[id], {sandboxId, cmd, data})
          --print("Events for "..id..": "..tostr(safecomm_queues[id]))
        end
      end
    end
    
    function safecomm_interface:handler(handler)
      table.insert(handlers, handler)
    end
    
    function safecomm_interface:removeHandler(handler)
      for i, h in ipairs(handlers) do
        if h == handler then
          table.remove(handlers, i)
          return
        end
      end
    end
    
    function safecomm_interface:listenOn(channel)
      safecomm_channels[channel] = safecomm_channels[channel] or {}
      table.insert(safecomm_channels[channel], sandboxId)
    end

    -- result is nil (=i still have stuff to do anyway), true (=wait for events) or socket list
    function safecomm_interface:yield(result, exitAfterEvents)
      exitAfterEvents = exitAfterEvents or false
      while true do
        --print("yield "..tostring(sleep))
        local events = coroutine.yield(result)
        --print("events after yield: "..tostr(events)..', handlers: '..tostring(#handlers))
        for e in elems(events) do
          for handler in elems(handlers) do
            handler(e[1], e[2], e[3])
          end
        end
        --print(sandboxId.." #events="..tostring(#events))
        if #events == 0 then break end
        if #events > 0 and exitAfterEvents then break end
      end
    end
    
    function safecomm_interface:daemon()
      while true do
        safecomm_interface:yield(true)
      end
    end
    
    -- this should probably go somewhere else
    function safecomm_interface:keepAlive(keepAlive)
      safecomm_keepAlive[sandboxId] = keepAlive or nil
    end
    
    -- like send, but waits for a result and returns that
    function safecomm_interface:invoke(...)
      local recp, cmd, data = safecomm_extract(...)
      local newdata = {recp, cmd}
      for k, v in pairs(data) do newdata[k] = v end
      newdata._answerId = 1
      safecomm_interface:send(newdata)
      local hasAnswer, answer = false, nil
      local handler = function(sender, cmd, data2)
        --print('answer handler, cmd: '..cmd)
        if cmd == 'answer' and data2.answerId == newdata._answerId then
          --print('hasAnswer=true')
          hasAnswer = true
          answer = data2.answer
        end
      end
      safecomm_interface:handler(handler)
      repeat
        safecomm_interface:yield(true, true)
        --print('hasAnswer: '..tostring(hasAnswer))
      until hasAnswer
      safecomm_interface:removeHandler(handler)
      --print("answer: "..tostring(answer))
      return answer
    end
    
    function safecomm_interface:answer(sender, data, answer)
      --print('data: '..tostr(data)..' from: '..sandboxId)
      if data._answerId ~= nil then
        safecomm_interface:send{sender, 'answer', answerId=data._answerId, answer=answer}
      end
    end
    
    return safecomm_interface
  end
  
  function cage:safecomm_hasEvents(id)
    events = safecomm_queues[id]
    return events ~= nil
  end
  
  function cage:safecomm_getEvents(id)
    events = safecomm_queues[id] or {}
    safecomm_queues[id] = nil
    return events
  end

  function cage:safecomm_send(from, to, cmd, data)
    safecomm_queues[to] = safecomm_queues[to] or {}
    table.insert(safecomm_queues[to], {from, cmd, data})
  end
  
  function cage:safecomm_registerSandbox(id)
    safecomm_channels[id] = {id}
  end
  
  function cage:shutdown()
    for _, info in pairs(cage.sandboxInfos) do
      if info.shutdownHandler then
        result = pcall(info.shutdownHandler)
        if not result then
          print('[warning: shutdown handler failed]')
        end
      end
    end
  end

  function cage:findRoutine(id)
    for r in elems(cage.routines) do
      if r.id == id then return r end
    end
    return nil
  end
  
  return cage
end
