-- lib_safelua.lua
-- the Safe Lua library
--
-- Version: Safe Lua 0.2 (http://safelua.sf.net)
-- by Stefan Reich (www.friedensrevolution.de)
-- Contact: stefan.reich.maker.of.eye@gmail.com
-- This code is public domain!

require "lib_datadumper"
require "lib_safecomm"

verbose = false
selectTimeout = 1

-- from stdlib
--- Find the number of elements in a table.
-- @param t table
-- @return number of elements in t
function size (t)
  local n = 0
  for _ in pairs (t) do
    n = n + 1
  end
  return n
end

function contains(l, e)
  for _, x in ipairs(l) do
    if x == e then return true end
  end
  return false
end

function info(msg)
  if verbose then
    io.stderr:write("["..msg.."]\n")
  end
end

function setTimeout(timeout)
  if timeout then
    local n
    if timeout >= 1000 then
      n = tostring(timeout/1000)..'k'
    else
      n = tostring(timeout)
    end
    debug.sethook(function() error('timeout <'..n..'>') end, "", timeout)
  end
end

function setHook(info, timeout, softtimeout)
  if timeout == 0 then timeout = nil end
  if softtimeout == 0 then softtimeout = nil end
  --print('timeout='..tostring(timeout)..', softtimeout='..tostring(softtimeout))
  if timeout == nil and softtimeout == nil then
    debug.sethook(function()
      info.instructioncounter = info.instructioncounter+1
      info.softinstructioncounter = info.softinstructioncounter+1
    end, '', 1000)
  else
    if timeout then timeout = timeout/1000 end
    if softtimeout then softtimeout = softtimeout/1000 end
    --debug.sethook(function() error('bla') end, 1)
    debug.sethook(function()
      info.instructioncounter = info.instructioncounter+1
      if timeout and info.instructioncounter >= timeout then
        error('timeout <'..tostring(timeout)..'k>')
      end
      info.softinstructioncounter = info.softinstructioncounter+1
      if softtimeout and info.softinstructioncounter >= softtimeout then
        error('timeout <'..tostring(softtimeout)..'k>')
      end
      --print("counters:", info.instructioncounter, " / ", info.softinstructioncounter)
    end, '', 1000)
  end
end

function safecall(user_code, env, timeout)
  setfenv(user_code, env)
  setTimeout(timeout)
  status, result = pcall(user_code)
  debug.sethook()
  return status, result
end

-- Make a basic sandbox. Nothing fancy and probably (!) safe. This will have
-- to be audited. We'll do that. Stuff will be bulletproof. I guarantee!
function makeSandbox(id, cage)
  local info = {instructioncounter=0, softinstructioncounter=0, output={}}
  cage.sandboxInfos[id] = info

  local env = {}
  
  -- the basics
  env.id = id
  env.print = print
  env.error = error
  env.dump = dump

  -- safe (we hope) parts of the library
  env.string = string
  env.tostring = tostring
  env.table = table
  env.ipairs = ipairs
  env.pairs = pairs
  env.type = type
  
  -- deny any fancy stuff (extensions)
  env.safelua = {}
  
  function env.safelua:requestExtension(name)
    error("Sorry. No extensions available (was requesting: "..name..")")
  end
  
  --[[function env.safelua:blockingOp(f)
    error("blockingOp is not allowed in this context")
  end]]
  
  function env.safelua:onShutdown(f)
    info.shutdownHandler = f
  end

  function env.safelua:exit()
    error("exit")
  end
  
  return env
end

-- Add safecomm facilities to sandbox
function addSafecomm(sandbox, cage)
  -- intercom
  sandbox.safecomm = cage:makeSafecommInterface(sandbox.id)
end

-- Is it bytecode? We don't do that. Source only.
function isBytecode(data)
  return string.byte(data) == 27
end

-- Load script from file or source-as-string (with a "** " prefix to distinguish)
-- Returns: flag, result/error
function loadScript(filename)
  local script
  if filename:sub(1, 3) == "** " then
    script = filename:sub(4)
  else
    info("Loading "..filename)      
    if not pcall(io.input, filename) then
      return false, "Script not found: "..filename
    end
    script = io.read("*all")
  end
  if isBytecode(script) then
    return false, "Loading bytecode is disallowed because it is not safe. Use actual source code."
  end
  code, err = loadstring(script, filename)
  if code == nil then
    return false, "Bad syntax - "..err
  end
  return true, code, script
end

-- Returns: flag, result/error
function runScript(filename, sandbox, timeout)
  status, code = loadScript(filename)
  if not status then return false, code end
  info("Running in sandbox")
  status, result = safecall(code, sandbox, timeout)
  return status, result
end

-- Returns: flag, result/error
function runCoroutine(filename, sandbox, timeout, softtimeout, sbinfo)
  local status, code, script = loadScript(filename)
  if not status then return false, code end
  sbinfo.code = script -- save the script so we can always find out what code is running
  info("Running in sandbox as coroutine")
  setfenv(code, sandbox)
  local function withdebug()
    setHook(sbinfo, timeout, softtimeout)
    code()
  end
  --thecoroutine = coroutine.create(function() return copcall(code) end)
  local thecoroutine = coroutine.create(withdebug)
  --print(coroutine.status(thecoroutine))
  return true, thecoroutine
end

function makeCoroutine(cage, i)
  local info = cage.sandboxInfos[cage.sandboxes[i].id]
  local status, routine = runCoroutine(cage.filenames[i], cage.sandboxes[i],
    info.timeout or cage.timeout,
    info.softtimeout or cage.softtimeout, info)
  if not status then
    io.stderr:write("Script error: "..tostring(routine).."\n")
    os.exit()
  end
  table.insert(cage.routines, {
    routine=routine,
    id=cage.sandboxes[i].id,
    sleeping=false,
    waitingOnSockets={}})
  cage:safecomm_registerSandbox(cage.sandboxes[i].id)
end

-- check all the sockets that scripts are waiting on
-- returns true if any sockets are waited on
-- or false if application should exit (no script is waiting on anything external)
function doSelect(cage, timeout)
  sockets = {}
  for routine in elems(cage.routines) do
    for s in elems(routine.waitingOnSockets) do
      table.insert(sockets, s)
    end
  end
  --print('#sockets: '..tostring(#sockets))
  if #sockets == 0 then
    return false
  end

  readers, writers = socket.select(sockets, {}, timeout)
  --print('#readers: '..tostring(#readers))

  if #readers > 0 then
    for routine in elems(cage.routines) do
      for s in elems(routine.waitingOnSockets) do
        if contains(readers, s) then
          routine.sleeping = false
          --print('system', routine.id, 'socketReady', nil)
          cage:safecomm_send('system', routine.id, 'socketReady', nil)
          break
        end
      end
    end
  end
  return true
end

function runCoroutines(cage)
  -- make coroutines
  cage.routines = {}
  for i=1, #(cage.filenames) do
    makeCoroutine(cage, i)
  end
  theRoutines = cage.routines
  
  doStuff = function()
    -- run coroutines
    local idleCycles = 0
    while #cage.routines > 0 do
      --doSelect(cage, selectTimeout)
      local i=1
      local anyNotSleeping = false
      while i <= #cage.routines do
        local r = cage.routines[i]
        if r then -- why is this not always true? I saw this fail oncve
          local id = r.id
          local hasEvents = cage:safecomm_hasEvents(id)
          --print(id.." hasEvents: "..tostring(hasEvents))
          if r.sleeping and hasEvents then
            r.sleeping = false
          end

          --dump(cage.safecomm_keepAlive)
          if cage.safecomm_keepAlive[id] then
            r.sleeping = false
          end

          if not r.sleeping then
            anyNotSleeping = true
            local events = cage:safecomm_getEvents(id)
            --print("events: "..#events)
            local status, result
            if coroutine.status(r.routine) == 'running' then
              print("Warning: coroutine "..r.id.." mysteriously in running state")
              status, result = true, true
            else
              --print("counter was: ", cage.sandboxInfos[id].softinstructioncounter)
              cage.sandboxInfos[id].softinstructioncounter = 0
              status, result = coroutine.resume(r.routine, events)
            end

            --print(status, result)
            if not status then
              io.stderr:write("Script error: "..tostring(result).."\n")
              cage.sandboxInfos[id].error = tostring(result)
            else
              if type(result) == 'table' then
                r.waitingOnSockets = result
              else
                r.waitingOnSockets = {}
              end
            end
            if coroutine.status(r.routine) == 'dead' then
              print("  "..id.." done ("..#cage.routines.." more to go)")
              --cage.sandboxInfos[id].ended = true
              table.remove(cage.routines, i)
              i = i-1
            else
              if result then
                print("routine "..tostring(i).." sleeps")
                r.sleeping = true
              end
            end
          end
        end
        
        i = i+1
      end
      if anyNotSleeping then
        idleCycles = 0
      else
        idleCycles = idleCycles + 1
        if idleCycles >= 2 then
          print("All scripts sleeping")
          if not doSelect(cage, selectTimeout) then
            break
          end
        end
      end
    end
  end
end
